import os
import sys
from datetime import datetime
from pathlib import Path
from t8s.log_config import LogConfig
from behave import given, when, then, model, step # type: ignore
from behave.model import Feature, Scenario # type: ignore
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL, log
from behave import use_step_matcher # type: ignore

from t8s.util import Util
from t8s.ts import TimeSerie
from t8s import get_sample_df

# use_step_matcher("re")
use_step_matcher("parse")

LogConfig().initialize_logger(INFO)
logger = LogConfig().getLogger()
print(f'\n\n• The script "environment.py" was loaded ...\n')
print(f'\n• The python version used is "{sys.version_info}"\n\n')

def list_files(prefix: str, context):
    l: list = Util.list_all_files(context.PARQUET_PATH)
    for item in l:
        logger.info(prefix + item)

def before_all(context):
    # Inicializa o contexto do behave
    context.status = 'initializing . . .'
    context.list_files = list_files
    logger.info(f'\n\nbefore_all() called . . .\n')
    logger.info(f'-------------------------------------------------')
    T8S_WORKSPACE_DIR = os.environ.get('T8S_WORKSPACE_DIR', '/Volumes/dev/t8s')
    CSV_PATH_STR:str = os.path.join(T8S_WORKSPACE_DIR, 'data', 'csv')
    CSV_PATH: Path = Path(CSV_PATH_STR)
    PARQUET_PATH_STR:str = os.path.join(T8S_WORKSPACE_DIR, 'data', 'parquet')
    PARQUET_PATH: Path = Path(PARQUET_PATH_STR)
    context.T8S_WORKSPACE_DIR = T8S_WORKSPACE_DIR
    context.CSV_PATH = CSV_PATH
    context.PARQUET_PATH = PARQUET_PATH
    logger.info(f'before_all: T8S_WORKSPACE_DIR = {context.T8S_WORKSPACE_DIR}')
    logger.info(f'before_all: CSV_PATH          = {context.CSV_PATH}')
    logger.info(f'before_all: PARQUET_PATH      = {context.PARQUET_PATH}')
    create_sample_ts_and_save_as_parquet(context)
    context.list_files(f'before_all: ', context)
    logger.info(f'-------------------------------------------------')

def write_ts_to_parquet_file(ts, parquet_path, filename: str):
    parquet_file_path_str: str = str(parquet_path) + '/' + filename
    path_ts = Path(parquet_file_path_str)
    # Devido a problemas de 'circular import' tivemos que usar a classe Util
    Util.to_parquet(ts, path_ts)

def create_sample_ts_and_save_as_parquet(context):
    start_timestamp = datetime(2022, 1, 1, 0, 0, 0)
    number_of_records = 4
    time_interval = 1 # hour
    dataframe, last_ts = get_sample_df(number_of_records, start_timestamp, time_interval)
    context.ts1 = TimeSerie(dataframe, format='wide', features_qty=len(dataframe.columns))
    # Grava a série temporal ts1 em parquet
    write_ts_to_parquet_file(context.ts1, context.PARQUET_PATH, 'ts_01.parquet')


def clean_data_dir(CSV_PATH, PARQUET_PATH):
    logger.info(f'clean_data_dir() called ...')
    for diretory in [CSV_PATH, PARQUET_PATH]:
        logger.info(f'clean_data_dir() -> diretory = {diretory}')
        for root, dirs, files in os.walk(diretory):
            for file in files:
                os.remove(os.path.join(root, file))
            for dir in dirs:
                os.rmdir(os.path.join(root, dir))
        for root, dirs, files in os.walk(diretory):
            for file in files:
                print(os.path.join(root, file))
            for dir in dirs:
                print(os.path.join(root, dir))
    return 'data directory empty'

# HOOK
def before_feature(context, feature: Feature):
    # Inicializa o status no contexto
    context.status = context.status if hasattr(context, 'status') else 'not defined'
    # Passo funções utilitárias via contexto para serem usadas nas steps.
    context.create_sample_ts_and_save_as_parquet = create_sample_ts_and_save_as_parquet
    # Execute um método uma única vez para todos os cenários de teste da Feature
    logger.info(f'-------------------------------------------------')
    logger.info(f'before_feature: T8S_WORKSPACE_DIR = {context.T8S_WORKSPACE_DIR}')
    logger.info(f'before_feature:  CSV_PATH         = {context.CSV_PATH}')
    logger.info(f'before_feature:  PARQUET_PATH     = {context.PARQUET_PATH}')
    logger.info(f'-------------------------------------------------')
    # A forma de passar estes dados para os steps seguintes é usando o objeto context

    # Atualiza o status no contexto
    if feature.filename == 'feature/01.create_timeserie.feature':
        pass

    logger.info(f'-------------------------------------------------')

# HOOK
def before_scenario(context, scenario: Scenario):
    logger.info(f'-------------------------------------------------')
    logger.info(f'before scenario: {scenario.name}')
    logger.info(f' context.status: {context.status}')
    logger.info(f'-------------------------------------------------')



# SPDX-FileCopyrightText: 2023-present João Antonio Ferreira <joao.parana@gmail.com>
#
# SPDX-License-Identifier: MIT

# import yaml
# from pathlib import Path
# from datetime import datetime
# from t8s.ts import TimeSerie



import os
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
from pandas import Series
from t8s.ts import TimeSerie
from t8s.ts_writer import TSWriter, WriteParquetFile
from t8s.log_config import LogConfig
from t8s.ts_builder import TSBuilder, ReadParquetFile # , ReadCsvFile

# to consider inf and -inf to be “NA” in computations, you can set:
pd.options.mode.use_inf_as_na = True

logger = LogConfig().getLogger()

class Util:
    @staticmethod
    def to_parquet(ts: TimeSerie, path_ts: Path):
        # def write_ts_to_parquet_file(ts, parquet_path, filename: str):
        logger.debug(f'Grava a série temporal (formato {ts.format}) em um arquivo parquet {path_ts}')
        ctx = TSWriter(WriteParquetFile())
        logger.debug("Client: Strategy was seted to write Parquet file.")
        ctx.write(Path(path_ts), ts)
        logger.debug(f'\nArquivo {str(path_ts)} gerado à partir da TimeSerie fornecida')

    @staticmethod
    def list_all_files(path: Path) -> list:
        assert path.exists() and path.is_dir()
        result = []
        for root, dirs, files in os.walk(path):
            for file in files:
                # logger.debug(os.path.join(root, file))
                result.append(os.path.join(root, file))
            for dir in dirs:
                # logger.debug(os.path.join(root, dir))
                result.append(os.path.join(root, dir))
        for item in result:
            logger.debug(item)

        return result

    @staticmethod
    def build_time_serie_from_parquet(workspace_dir: Path, filename: str) -> TimeSerie:
        DATA_PATH = os.path.join(workspace_dir, 'data')
        PARQUET_DATA_PATH = os.path.join(DATA_PATH, 'parquet')
        parquet_input_file = os.path.join(PARQUET_DATA_PATH, filename)
        logger.debug(f'parquet_input_file: {parquet_input_file}, {type(parquet_input_file)}')
        path = Path(parquet_input_file)
        logger.debug(f'{path},  {type(path)}')
        ts = TimeSerie.empty()
        assert isinstance(path, Path), "path must be a Path object"
        if (str(path)).endswith('.parquet'):
            context = TSBuilder(ReadParquetFile())
            logger.debug("Client: ReadStrategy is set to read Parquet file.")
            ts = context.build_from_file(path)
        return ts

    # Função para obter a quantidade de valores nulos e não nulos consecutivos
    #
    # O resultado é um Dataframe com duas colunas, a primeira com os valores True e False, indicando
    # se o valor é nulo ou não, e a segunda com a quantidade de valores consecutivos iguais.
    # Valor: True se o valor é nulo, False se não é nulo
    @staticmethod
    def get_null_and_notnull_consecutive_counts(df_to_process: pd.DataFrame, col_name:str) -> pd.DataFrame:
        df_ret = df_to_process[col_name].isna()
        # print('df_ret:\n',  df_ret)
        # Obtendo os grupos consecutivos de valores iguais
        groups = df_ret.ne(df_ret.shift()).cumsum()
        inicio = np.where(groups.ne(groups.shift()))
        # Contando a quantidade de elementos em cada grupo
        counts = groups.value_counts()
        # Obtendo os valores dos grupos
        group_values = df_ret.groupby(groups).first()
        # Juntando as informações em um único DataFrame
        result = pd.DataFrame({'Value': group_values, 'Counts': counts,'Start': inicio[0]})
        result.reset_index(drop=True, inplace=True)
        return result

    @staticmethod
    def get_numeric_column_names(df) -> list:
        ret:list = []
        for idx, c in enumerate(df.columns):
            if (df[c].dtype == float or df[c].dtype == int or
            df[c].dtype == np.float64 or df[c].dtype == np.int64 or
            df[c].dtype == np.float32 or df[c].dtype == np.int32):
                ret.append(c)
        ret.sort()
        return ret

    @staticmethod
    def identify_start_and_end_of_nan_block(df, col) -> pd.DataFrame | None:
        logger.debug(f'\ncol: {col}, \ndf:\n {df[col]} \ntype: {type(df[col])}')
        # O schema da saida é: {'Value': [bool array], 'Counts': [int array], 'Start': [int array]}
        logger.debug(f'{df[col].head(1)}, {df[col].head(1)}, {type(df[col].head(1))}')
        x: Series = df[col]
        # if df[col].head(1).isnull().values.any():
        if np.isnan(df[col].iloc[1]):
            logger.debug('Primeiro elemento da série é NaN')
            return None
        else:
            logger.debug('Primeiro elemento da série não é NaN')
            null_and_notnull_counts = Util.get_null_and_notnull_consecutive_counts(df, col)
            return null_and_notnull_counts

    @staticmethod
    def identify_all_start_and_end_of_nan_block(df) -> tuple[dict[str, pd.DataFrame], int]:
        last_idx = 0
        ret_all_s_e_nan_block = {}
        cols = Util.get_numeric_column_names(df)
        for idx, col in enumerate(cols):
            key = f'{idx}-{col}'
            l = Util.identify_start_and_end_of_nan_block(df, col)
            ret_all_s_e_nan_block.update({key: l})
            last_idx = idx

        return (ret_all_s_e_nan_block, last_idx)



import os
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
from t8s import get_sample_df
from t8s.log_config import LogConfig
from t8s.util import Util
from t8s.io import IO
from t8s.ts import TimeSerie
from t8s.ts_writer import TSWriter, WriteParquetFile
from behave import given, when, then, use_step_matcher, step # type: ignore
from behave.model import Table
from behave_pandas import table_to_dataframe, dataframe_to_table
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL

LogConfig().initialize_logger(DEBUG)
logger = LogConfig().getLogger()

"""
Feature: Create a time series set using Dataframe, CSV and Parquet

Value Statement:
  As a data analyst
  I want the hability to create a timeseries set using Dataframe, CSV and Parquet
  So that I can start studying the data right away and propose solutions for the business.

  Background:
    Given that I have a T8S_WORKSPACE_DIR and a bunch of CSV and Parquet files to analise
"""

@given(u'that I have a T8S_WORKSPACE_DIR and a bunch of CSV and Parquet files to analyze')
def setup(context):
    logger.info(f'-------------------------------------------------')
    logger.info(f'Background @given: T8S_WORKSPACE_DIR = {context.T8S_WORKSPACE_DIR}')
    logger.info(f'Background@given:  CSV_PATH = {context.CSV_PATH}')
    logger.info(f'Background@given:  PARQUET_PATH = {context.PARQUET_PATH}')
    logger.info(f'-------------------------------------------------')
    # A forma de passar estes dados para os steps seguintes é usando o objeto context

"""
  Scenario: Create 2 time series with sample data using Dataframes and save in CSV files at T8S_WORKSPACE/data/csv directory
    Given a start timestamp, a number of records and a time interval
    When I create 2 time series using Dataframes with sample data
    Then I have a time series with the `correct` number of rows and columns, schema and time interval
    And I have a CSV file in T8S_WORKSPACE/data/csv correctelly formated
    # Constraint: The first  Dataframe doesn't have nulls or invalid values, but the second does
"""

@given(u'a start timestamp, a number of records and a time interval')
def create_dataframe(context):
    start_timestamp = datetime(2022, 1, 1, 0, 0, 0)
    number_of_records = 4
    time_interval = 1 # hour
    logger.info(f'@given: -> Building a sample Dataframe')
    dataframe1, last_ts = get_sample_df(number_of_records, start_timestamp, time_interval)
    context.time_interval = time_interval
    context.number_of_records = number_of_records
    context.dataframe1 = dataframe1
    context.last_ts = last_ts
    # altera a série temporal incluindo NaN e inválidos
    dataframe2, last_ts = get_sample_df(number_of_records, last_ts + pd.Timedelta(hours=1), time_interval)
    dataframe2_str = ''
    logger.debug(f'Original: {dataframe2_str}')
    # altera a série temporal incluindo NaN e inválidos
    dataframe2.iloc[0, 1] = 'Missing'
    dataframe2.iloc[1, 1] = np.nan
    dataframe2.iloc[2, 2] = np.nan # ATENÇÃO: ao atribuir NaN, o Pandas converte o tipo da coluna de int32 para float64
    dataframe2.iloc[3, 1] = 'Bad'
    context.dataframe2 = dataframe2
    try:
        dataframe2_str = str(dataframe2)
    except Exception as e:
        logger.error(f'@given: -> dataframe2_str = str(dataframe2) -> {e}')

    context.last_ts = last_ts
    logger.info(f'@given: -> CSV_PATH = {context.CSV_PATH}')

@when(u'I create 2 time series using Dataframes with sample data')
def create_time_series(context):
    logger.info(f'@when: => CSV_PATH = {context.CSV_PATH}')
    # logger.info(f'@when: => context.dataframe1 = \n{context.dataframe1}')
    # logger.info(f'@when: => context.dataframe2 = \n{context.dataframe2}')
    # Gerando a primeira série temporal
    df1 = context.dataframe1
    context.ts1 = TimeSerie(df1, format='wide', features_qty=len(df1.columns))
    df2 = context.dataframe2
    context.ts2 = TimeSerie(df2, format='wide', features_qty=len(df2.columns))

@then(u'I have a time series with the `correct` number of rows and columns, schema and time interval')
def check_schema(context):
    logger.info(f'@then: -> CSV_PATH = {context.CSV_PATH}')
    logger.info(f'@then: -> context.ts1 = \n{context.ts1}')
    logger.info(f'@then: -> context.ts2 = \n{context.ts2}')
    logger.info('------------------------------------------------------------')
    def check_time_interval_equals(ctx, value):
        last_ts = None
        for idx, row in enumerate(ctx.dataframe1.itertuples()):
            if idx == 0:
                last_ts = row[1]
            else:
                interval: pd.Timedelta = row[1] - last_ts
                interval_seconds:int = interval.seconds
                time_interval = interval_seconds / 3600
                assert time_interval == value
                last_ts = row[1]
        logger.info(f'@then: -> time_interval checked')
    check_time_interval_equals(context, context.time_interval)
    assert len(context.ts1.df) == context.number_of_records, 'context.ts1.df must have the correct number of records'
    assert len(context.ts2.df) == context.number_of_records, 'context.ts2.df must have the correct number of records'

@then(u'I have a CSV file in T8S_WORKSPACE/data/csv correctelly formated')
def save_csv(context):
    logger.info(f'@then: => CSV_PATH = {context.CSV_PATH}')
    logger.info(f'@then: => len(context.dataframe1) = \n{len(context.dataframe1)}')
    csv_file_path_str_1 = str(context.CSV_PATH) + '/ts_01.csv'
    IO.dataframe_to_csv_file(context.dataframe1, Path(csv_file_path_str_1))
    csv_file_path_str_2 = str(context.CSV_PATH) + '/ts_02.csv'
    IO.dataframe_to_csv_file(context.dataframe2, Path(csv_file_path_str_2))

@then(u'I have a Parquet file in T8S_WORKSPACE/data/parquet correctelly formated with metadata annotations')
def save_parquet(context):
    logger.info(f'@then: => PARQUET_PATH = {context.PARQUET_PATH}')
    logger.info(f'@then: => len(context.dataframe1) = \n{len(context.dataframe1)}')
    logger.info(f'@then: => len(context.dataframe2) = \n{len(context.dataframe2)}')

    def write_ts_to_parquet_file(ts, parquet_path, filename: str):
        parquet_file_path_str: str = str(parquet_path) + '/' + filename
        path_ts = Path(parquet_file_path_str)
        # Devido a problemas de 'circular import' tivemos que usar a classe Util
        Util.to_parquet(ts, path_ts)

    # Grava a série temporal ts1 em parquet
    write_ts_to_parquet_file(context.ts1, context.PARQUET_PATH, 'ts_01.parquet')

    # Grava a série temporal ts2 em parquet
    # write_ts_to_parquet_file(context.ts2, context.PARQUET_PATH, 'ts_02.parquet')

@then(u'I have a text representation for the first time serie like this below')
def check_text(context):
    logger.debug(f'check_text : context.text -> \n{str(context.text)}')
    logger.debug(f'\ncheck_text : context.ts1 -> \n{str(context.ts1)}')
    def assert_line_by_line(context):
        ts1_line = str(context.ts1).splitlines()
        for idx, line in enumerate(context.text.splitlines()):
            logger.info(f'check_text : line {idx} -> \n"{line}" == "{ts1_line[idx]}"')
            msg =f'context.text must be equal to str(ts1). Failed in line {idx} -> "{line}" != "{ts1_line[idx]}"'
            assert line.strip() == ts1_line[idx].strip(), msg
    assert_line_by_line(context)
    logger.info('\ncheck_text passed !')

# -----------------------------------------------------------------------------------------------

"""
  Scenario: Second, I create 1 time series at T8S_WORKSPACE/data directory using a literal sample data
    Given a the table below as input
      | datetime            | float       | float       |
      | timestamp           | temperatura | velocidade  |
      | 2022-01-01 00:00:00 | 3.0         | 30.0        |
      | 2022-01-01 01:00:00 | 4.1         | 1124        |
      | 2022-01-01 02:00:00 | 5.2         | 3276        |
    When converted to a data frame using 1 row as column names and 1 column as index
    And printed using data_frame_to_table function
    Then I build a time series with the `correct` number of rows and columns
"""

use_step_matcher("parse")

@given(u'a the table below as input')
def the_table(context):
    table: Table = context.table
    context.input = table
    table.headings
    logger.info('the_table() called ...\n' + str(context.input) + ' ->' + str(type(context.input)))
    logger.info(f'{table.headings}')
    for row in table.rows:
        logger.info(f'{row}')
    logger.info(f'the_table() passed !')

@when(u'converted to a data frame using {column_levels:d} row as column names and {index_levels:d} column as index')
def to_dataframe(context, column_levels, index_levels):
    parsed: pd.DataFrame = table_to_dataframe(context.input, column_levels=column_levels)
    for idx, col in enumerate(parsed.columns):
        logger.info(f'{idx}\t-> {col}, {type(parsed[col])}, {type(parsed[col][0])}')
    context.parsed = parsed
    logger.info(f'context.parsed =\n{context.parsed}\nType of context.parsed =\n{type(context.parsed)}')
    logger.info(f'context.parsed.columns = {context.parsed.columns}')
    logger.info(f'Test of to_dataframe passed !')

@when(u'printed using data_frame_to_table function')
def print_to_table(context):
    table = dataframe_to_table(context.parsed)
    assert isinstance(table, str)
    logger.info(f'table =\n{table}\nType of table = {type(table)}')
    logger.info(f'Test of print_to_table passed !')

@then(u'I build a time series with the `correct` number of rows and columns')
def table_to_ts(context):
    df3: pd.DataFrame = context.parsed
    logger.info(f'parsed =\n{df3}\nType of parsed = {type(df3)}')
    context.ts3 = TimeSerie(df3, format='wide', features_qty=len(df3.columns))
    logger.info(f'context.ts3 =\n{context.ts3}')
    logger.info(f'Test of table_to_ts passed !')



# -*- coding: utf-8 -*-

# from __future__ import annotations usado apenas nas versões anteriores a 3.7
from abc import ABC, abstractmethod
from logging import captureWarnings
from typing import Any, Optional
from pathlib import Path
from datetime import datetime
from t8s.ts import TimeSerie
from t8s.log_config import LogConfig
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

logger = LogConfig().getLogger()


class Strategy(ABC):
    """
    The Strategy interface declares operations common to all supported versions
    of some algorithm.

    The Context uses this interface to call the algorithm defined by Concrete
    Strategies.
    """

    @abstractmethod
    def do_write(self, data: Any, ts: Any) -> Any:
        pass


"""
Concrete Strategies implement the algorithm while following the base Strategy
interface. The interface makes them interchangeable in the Context.
"""


class TSWriter:
    """
    The Context defines the interface of interest to clients.
    """

    def __init__(self, strategy: Strategy) -> None:
        """
        Usually, the Context accepts a strategy through the constructor, but
        also provides a setter to change it at runtime.
        """
        logger.info('TSBuilder constructor received strategy: %s', strategy)
        self._strategy = strategy

    @property
    def strategy(self) -> Strategy:
        """
        The Context maintains a reference to one of the Strategy objects. The
        Context does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface.
        """

        return self._strategy

    @strategy.setter
    def strategy(self, strategy: Strategy) -> None:
        """
        Usually, the Context allows replacing a Strategy object at runtime.
        """

        self._strategy = strategy

    def write(self, path: Path, ts: TimeSerie) -> None:
        """
        The Context delegates some work to the Strategy object instead of
        implementing multiple versions of the algorithm on its own.
        """

        # ...

        logger.info("Context: write")
        result = self._strategy.do_write(path, ts)
        # logger.info('result type from do_write() is: ' + str(type(result)))

        # ...
        return result


class WriteParquetFile(Strategy):
    # O método do_write é VOID, pois o resultado é gravado em disco. Caso ocorra
    # algum problema uma `Exception` é lançada.
    def do_write(self, path: Path, ts: TimeSerie) -> None:
        logger.info('Using WriteParquetFile strategy')
        # Grava os dados em formato Parquet com metadados do objeto TimeSerie
        # to_parquet(path, df, self.format, self.features)
        table = pa.Table.from_pandas(ts.df)
        # table = table.replace_schema_metadata({'format': self.format, 'features': self.features})
        table = table.replace_schema_metadata(
            {b'format': str(ts.format).encode(), b'features': str(ts.features).encode()}
        )
        result = pq.write_table(table, path)
        return result


class WriteCsvFile(Strategy):
    # O método do_write é VOID, pois o resultado é gravado em disco. Caso ocorra
    # algum problema uma `Exception` é lançada.
    def do_write(self, path: Path, ts: TimeSerie) -> None:
        logger.info('Using WriteCsvFile strategy')
        return None


if __name__ == "__main__":
    from logging import INFO, DEBUG
    import numpy as np
    import pandas as pd
    import subprocess

    LogConfig().initialize_logger(DEBUG)
    path_str: str = 'ts_01.parquet'
    path = Path(path_str)
    # Cria uma série temporal multivariada com três atributos: timestamp, temperatura e velocidade
    data = {
        'timestamp': [
            datetime(2022, 1, 1, 0, 0, 0),
            datetime(2022, 1, 1, 1, 0, 0),
            datetime(2022, 1, 1, 2, 0, 0),
            datetime(2022, 1, 1, 3, 0, 0),
        ],
        'temperatura': np.array([25.0, 26.0, 27.0, 23.2], dtype=np.float32),
        'velocidade': np.array([2000, 1100, 1200, 4000], dtype=np.int32),
    }
    # Cria uma série temporal multivariada
    ts = TimeSerie(data, format='wide', features_qty=3)
    # Grava a série temporal em parquet
    print(f'Grava a série temporal (formato {ts.format}) em um arquivo parquet {path}')
    (TSWriter(WriteParquetFile())).write(Path(path_str), ts)
    print(f'Arquivo {str(path)} gerado à partir da TimeSerie fornecida')
    #  Supondo o arquivo /usr/local/bin/lsla existindo no sistema com o conteudo abaixo:
    #  #!/bin/bash
    #
    #  ls -lA *.parquet
    #
    result = subprocess.run(['lsla'], capture_output=True, text=True)
    print(result.stdout)



import types
import yaml
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
from t8s import get_sample_df
from t8s.ts import TimeSerie
from t8s.ts_writer import TSWriter, WriteParquetFile
import pyarrow.parquet as pq         # type: ignore


from t8s.log_config import LogConfig
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL
logger = LogConfig().getLogger()

def check_schema(ts, parquet_file: Path, types: list[type]):
    """
    Verifica o schema do arquivo parquet gerado e os tipos dos dados de cada coluna
    contra os tipos esperados encontrados no Dataframe que é parte da TimeSerie.
    """
    logger.debug(f'Verificando os tipos dos dados gravados no arquivo parquet {parquet_file} contra os tipos esperados:', types)
    # Lê os metadados do arquivo Parquet
    metadata: pq.FileMetaData = pq.read_metadata(parquet_file)
    logger.debug('\nParquet file metadata:\n' + str(metadata.to_dict()) + '\n' + str(metadata.metadata))
    assert isinstance(metadata, pq.FileMetaData), "metadata must be a pq.FileMetaData object"
    dict_meta: dict = metadata.to_dict()
    logger.debug('\n-------------------------------')
    logger.debug('created_by: ' + str(dict_meta['created_by']))
    # Imprime o valor do metadado 'format'
    format = metadata.metadata[b'format'].decode()
    features = metadata.metadata[b'features'].decode()
    logger.info('format: ' + format + ' type(format) ' + str(type(format)))
    logger.info('features: ' + str(features) + ' type(features): ' + str(type(features)))
    assert isinstance(format, str), "format metadada must be a string"
    assert isinstance(features, str), "features metadada must be a string"
    features_qty = int(features)
    assert features_qty > 1, "features_qty must be greater than one"
    # logger.debug('format', dict_meta['format'])
    # logger.debug('features', dict_meta['features'])
    # Imprime o esquema do arquivo Parquet
    logger.debug('\ntype(metadata.schema): ' + str(type(metadata.schema)) + '\t' + str(metadata.schema))
    # Imprime as colunas do arquivo Parquet
    # logger.debug('metadata.column_names', metadata.column_names)
    # Imprime as estatísticas do arquivo Parquet
    physical_types = []
    if metadata.num_row_groups > 0:
        for idx in range(metadata.num_columns):
            stats = metadata.row_group(0).column(idx).statistics
            logger.debug(stats)
            logger.debug('-----------------------------')
            logger.info('physical_type = ' + stats.physical_type)
            physical_types.append(stats.physical_type)
    # Exemplo: INT64, FLOAT, INT32 -> int64(datetime64[ns]), float32, int32
    logger.debug('Parquet physical_types:', ', '.join(physical_types))
    logger.debug(ts.df.info())
    logger.debug('TimeSerie types expected:', ', '.join([str(t) for t in types]))
    pass

def test_to_parquet():
    TimeSerie.empty()
    # Cria uma série temporal multivariada com três atributos: timestamp, temperatura e velocidade

    number_of_records = 4
    time_interval = 1 # hour
    df, last_ts = get_sample_df(number_of_records, datetime(2022, 1, 1, 0, 0, 0), time_interval)
    columns_types = [type(df[col][0]) for col in df.columns]
    logger.debug('columns_types =', columns_types)
    # data = df.to_dict()
    ts = TimeSerie(df, format='wide', features_qty=len(df.columns))
    cols_str = [name for name in sorted(df.columns)]
    logger.debug('cols_str :', ', '.join(cols_str))
    logger.debug(f'Dataframe com {len(df.columns)} colunas: {df.columns}')
    # Imprime a série temporal
    logger.debug(ts)
    # Grava a série temporal em parquet
    path_str: str = 'data/parquet/ts_01.parquet'
    path = Path(path_str)
    logger.debug(f'Grava a série temporal (formato {ts.format}) em um arquivo parquet {path}')
    ctx = TSWriter(WriteParquetFile())
    logger.debug("Client: Strategy was seted to write Parquet file.")
    ctx.write(Path(path_str), ts)
    logger.debug(f'\nArquivo {str(path)} gerado à partir da TimeSerie fornecida')
    check_schema(ts, path, [datetime, np.float32, np.int32])
    logger.info('FIM')

if __name__ == "__main__":
    LogConfig().initialize_logger(DEBUG, log_file = 'tests/to_parquet.log')
    test_to_parquet()



# -*- coding: utf-8 -*-

from pathlib import Path
from datetime import date, datetime
import gzip
import csv
import os
import gc
import yaml
from abc import ABC, abstractmethod
from typing import Any, Optional, TypeVar, TypeAlias

import numpy as np
import pandas as pd
import pyarrow as pa         # type: ignore
import pyarrow.parquet as pq # type: ignore

from t8s.log_config import LogConfig

logger = LogConfig().getLogger()

import re
from enum import Enum
from t8s import get_numeric_regex
from t8s import PathStr

NULL_PATH: Path = Path('/NullPathObject')


class IO:
    all_columns_with_errors: dict[str, str] = {}
    all_tokens_with_errors: dict[str, str] = {}
    @staticmethod
    def add_column_name_and_invalid_token(column: str, token: str) -> None:
        # Atualize o dicionário xpto
        IO.all_columns_with_errors[column] = ''
        IO.all_tokens_with_errors[token] = ''

    # A linguagem Python não tem um meio de definir um método privado, mas o uso de dois underscore
    # no inicio do nome do método é uma convenção que indica que o método é privado e que estamos
    # usando a técnica name mangling para impedir a visibilidade padrão. Portanto o método
    # __check_csv_schema não deve ser chamado diretamente de fora da classe.
    @staticmethod
    def __check_csv_schema(csv_file_list: list[Path]):
        start_at = datetime.now()
        logger.debug('\nAtenção: Links simbólicos serão resolvidos, caso existam.')
        for f in csv_file_list:
            logger.debug('', f.resolve())

        csv_file_schema = ''
        # logger.debug('csv_file_list:', csv_file_list)
        first_lines = []
        for f in csv_file_list:
            with open(f) as f:
                first_line = f.readline()
                first_lines.append(first_line)

        # logger.debug('first_lines:', first_lines)
        for i, line in enumerate(first_lines):
            if line != first_lines[0]:
                msg = (
                    f"Schema do arquivo {csv_file_list[i]} não é igual "
                    + "ao do arquivo {csv_file_list[0]} que é {first_lines[0]}"
                )
                raise ValueError(msg)

        if len(first_lines) > 0:
            csv_file_schema = first_lines[0]
            logger.debug('schema for CSV file: %s', csv_file_schema)

        # Podemos usar o Pandas para verificar os valores de cada coluna para encontrar possíveis
        # valores inválidos que não podem ser convertidos para o tipo da coluna. Por exemplo,
        # valores como  'Configure', 'Bad' e 'NotNumber', por exemplo, devem ser substituidos
        # por 'NaN' (np.na do numpy) para que possamos fazer a conversão para float.
        # Observe que o Pandas já tenta fazer uma inferencia de tipo para cada coluna, mas
        # quando não consegue determinar o tipo de uma coluna com precisão e sem ambiguidades
        # ele assume o tipo object. Assim, se o tipo de uma coluna for object, devemos usar uma
        # heuristica para determinar o tipo correto. Por exemplo, se a coluna contém apenas números
        # inteiros, podemos converter para int32 ou int64, dependendo do tamanho do número. Se a
        # coluna contém apenas números de ponto flutuante, podemos converter para float32 ou float64,
        # dependendo do tamanho do número. Se a coluna contém apenas strings, podemos converter para
        # string. Se a coluna contém apenas datas, podemos converter para datetime64[ns]. Se a coluna
        # contém apenas valores booleanos, podemos converter para bool. Se a coluna contém apenas NaN,
        # podemos converter para np.nan. Se a coluna contém apenas valores inválidos, podemos converter
        # para np.nan. Se a coluna contém apenas valores inválidos e NaN, podemos converter para np.nan.
        types: list[type] = []
        type_map: dict[str, type] = {}
        first_types = []
        first_map = {}
        for idx, csv_file in enumerate(csv_file_list):
            types, type_map = IO.check_data_types_in_csv_file(str(csv_file))
            if idx == 0:
                first_types = types
                first_map = type_map
            else:
                if types != first_types:
                    msg = (
                        f"Os tipos de dados do arquivo {csv_file} não são iguais "
                        + "aos tipos de dados do arquivo {csv_file_list[0]}"
                    )
                    raise ValueError(msg)

        if len(IO.all_columns_with_errors.keys()) > 0:
            logger.debug('Erros de tokens inválidos encontrados em campos numéricos dos arquivos CSV')
            for error in IO.all_columns_with_errors.keys():
                logger.debug(f'{error}')
        logger.debug(f'Este método __check_csv_schema() demorou: {datetime.now()-start_at}')
        return (types, type_map)

    @staticmethod
    def get_column_names_and_types_from_csv_file(csv_file: PathStr) -> tuple[list[str], list[type], dict[str, type]]:
        # Define os tipos de dados das colunas assumindo a primeira como pd.Timestamp e a ultima como target
        # e todas as outras como features numéricas. Outros caso de uso não são tratados no momento.
        # Verifica se é arquivo CSV no formato texto, sem compactação e com a extensão .csv
        if not os.path.isfile(csv_file):
            raise ValueError(f"O arquivo {csv_file} não existe")

        column_names = []
        first_line = ''

        # Trata gzip compressed data
        if csv_file.endswith('.csv.gz'):
            # Abra o arquivo .txt.gz e leia a primeira linha como uma string
            with gzip.open(csv_file, 'rt') as f:
                first_line = f.readline()
            column_names = first_line.split(',')
        # Trata CSV não comprimido
        if csv_file.endswith('.csv'):
            with open(csv_file) as f:
                first_line = f.readline()
            column_names = first_line.split(',')
        column_type_dict: dict[str, type] = {}
        column_type_list: list[type] = []
        last_index = len(column_names) - 1
        for idx, col_name in enumerate(column_names):
            if idx == 0:
                column_type_list.append(datetime)
                column_type_dict[col_name] = pd.Timestamp
            elif idx == last_index:
                column_type_list.append(str)
                column_type_dict[col_name] = str
            else:
                column_type_list.append(np.float32)
                column_type_dict[col_name] = np.float32
            logger.debug(f'{idx}\t"{col_name}": type: {column_type_list[idx]}')

        return column_names, column_type_list, column_type_dict

    @staticmethod
    def check_data_types_in_csv_file(csv_file: PathStr) -> tuple[list[type], dict[str, type]]:
        # TODO: No caso em que a primeira linha do CSV não tenha a definição dos nomes das colunas
        #       podemos usar o parâmetro header=None para que o Pandas crie nomes para as colunas.
        # -----------------------------------------------------------------------------------------------
        column_names, column_type_list, column_type_dict = IO.get_column_names_and_types_from_csv_file(csv_file)

        # Lê o arquivo CSV
        nan_values_tokens = ['Bad', 'Missing', 'Configure', 'nan', 'NaN']
        parse_dates: list[int] = [1] # Assumindo que a coluna 1 é a coluna de timestamp
        df = pd.read_csv(Path(csv_file)) # , parse_dates=parse_dates, na_values=nan_values_tokens

        # Valida os tipos de dados
        def check_number_regex(x: Any) -> bool:
            regex = get_numeric_regex()
            if not re.match(regex, str(x)):
                # logger.debug(f'check_number_regex() -> parms= col_name: {col_name}, token: {x}')
                IO.add_column_name_and_invalid_token(col_name, str(x))
                return False
            return True

        for i, col_type in enumerate(column_type_list):
            col_name = column_names[i]
            if col_name != '' and (col_type == np.float32 or col_type == np.float64):
                numeric_values = df[col_name].apply(check_number_regex).all()
                # logger.debug(df[col_name])
                if not numeric_values:
                    logger.error(f"A coluna {col_name} não contém apenas valores do tipo {col_type}")

        # Delete the used DataFrame
        del df

        # Perform garbage collection
        gc.collect()
        return (column_type_list, column_type_dict)

    # Quando directory, que é um Path, é informado ordenamos a lista de nomes
    # de arquivos CSV por convenção. O método retorna um Boolean que indica se
    # o schema de todos os arquivos CSV é o mesmo.
    @staticmethod
    def check_schema_in_csv_files(directory: Path = NULL_PATH, csv_files: list[Path] = []) -> bool:
        result = False
        csv_file_list = []
        if directory == NULL_PATH and len(csv_files) == 0:
            msg = 'Você deve fornecer um diretório ou uma lista de arquivos'
            raise ValueError('Parâmetros inválidos')
        elif directory != NULL_PATH and len(csv_files) > 0:
            msg = 'Você deve escolher entre  fornecer um diretório ou ' + 'uma lista de arquivos, mas não os dois !'
            raise ValueError('Parâmetros ambíguos. ' + msg)
        elif len(csv_files) > 0:
            csv_file_list = csv_files
        else:
            # Quando directory, que é um Path, é informado ordenamos a
            # lista de nomes de arquivos CSV por convenção
            csv_file_list = list(directory.glob('*.csv'))
            csv_file_list.sort()

        if len(csv_file_list) == 0:
            result = False
            raise ValueError(f"Não existem aquivos CSV no diretório")

        IO.__check_csv_schema(csv_file_list)
        result = True

        return result

    # Quando estamos lendo de arquivos CSV podem aparecer valores que não tem uma
    # representação valida para NaN e portanto atrapalha o processo de inferência
    # de tipo do Pandas. Assim tokens como  'Configure', 'Bad' e 'NotNamber',
    # por exemplo, devem ser substituidos por NaN (np.na do numpy) para que possamos
    # fazer a conversão para float. O método check_convertion_to_float() faz isso.
    @staticmethod
    def __check_convertion_to_float(col_name:str, cols: pd.Series) -> tuple[str, list[str]]:
        assert isinstance(cols, pd.Series)
        not_float_values = []
        # Definindo a expressão regular para identificar números de ponto flutuante
        # padrao = r'[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?'
        padrao = get_numeric_regex()
        # Use o método items() para iterar sobre os itens da série
        for idx, valor in cols.items():
            # logger.debug(idx, valor)
            if (not re.match(padrao, str(valor))) and (not pd.isna(valor)):
                if isinstance(valor, str):
                    if valor not in ['NaN', '-NaN', '-nan', 'nan', '<NA>', 'N/A',
                                     'NA', 'NULL', 'None', 'n/a', 'null']:
                        # logger.debug(idx, '\t', valor, 'não é uma representação valida para NaN')
                        not_float_values.append(valor)
                else:
                    # logger.debug(idx, '\t', valor, 'não e um número valido e nem uma representação valida para NaN')
                    not_float_values.append(valor)
        if len(not_float_values) > 0:
            logger.debug('not_float_values: ' + str(not_float_values) + ' for column: ' + str(col_name))

        return (col_name, not_float_values)

    @staticmethod
    def check_convertion_to_float(numeric_columns: list[str], df: pd.DataFrame) -> dict[str, list[str]]:
        all_problems = {}
        for col in numeric_columns:
            logger.debug('Verificando a conversão para tipo Number da coluna', col)
            col_name, not_float_values = IO.__check_convertion_to_float(col, df[col])
            if len(not_float_values) > 0:
                all_problems.update({col_name: not_float_values})
        return all_problems

    # path é um Path identificando um nome de arquivo CSV e a extensão deve
    # ser obrigatoriamente '.csv' (em minusculo). Isto é uma convenção.
    # Além disso o arquivo deve existir e possuir na primira linha a lista
    # de nomes das colunas / campos (schema). A quantidade de elementos na
    # lista 'column_types' deve ser igual a quntidade de colunas no arquivo,
    # ou seja, deve ser igual ao tamanho da lista de nomes especificada na
    # primeira coluna do arquivo CSV. Este método verifica os valores dos
    # campos numéricos contra as regras para Not A Number e retorna um
    # Dataframe com os tipos de dados numéricos corretos para cada coluna.
    @staticmethod
    def read_csv_file(path: Path, column_types: list[type],
            nan_values:list[str] = [], parse_dates: list[int] = [1],
            date_format:str = 'AAAA-MM-DD HH:MM:SS') -> pd.DataFrame:
        for value in column_types:
            assert isinstance(value, type), f'Erro: {value} não é um tipo válido'

        # Define o conversor para cada tipo de coluna
        # Crio um dicionário com os conversores e tipos de colunas
        converters_tmp = {
            f'{i}': column_types[i] for i in range(len(column_types))
        }

        # logger.debug('converters:', converters_tmp, type(converters_tmp))
        result: pd.DataFrame = pd.read_csv(filepath_or_buffer = path, sep=',', delimiter=None,
            header='infer', dtype=None, engine='c', converters=converters_tmp,
            na_values=nan_values, keep_default_na=True, na_filter=True, verbose=True,
            skip_blank_lines=True, parse_dates=parse_dates, date_format=date_format,
            dayfirst=False, cache_dates=False, compression='infer', thousands=None,
            decimal='.', quotechar='"', # Not Supported: lineterminator=os.linesep
            quoting=csv.QUOTE_MINIMAL, doublequote=True, comment=None, encoding='utf-8',
            encoding_errors='strict', on_bad_lines='error', delim_whitespace=False,
            low_memory=True, memory_map=False, dtype_backend='numpy_nullable')
        logger.debug(result.info())
        # Criando um dicionário com os conversores de tipos para as colunas
        converters =  {}
        if len(converters_tmp) > 0:
            # TODO: Verificar como converter a primeira coluna que é Timestamp e é index.
            for idx, col in enumerate(result.columns):
                converters[col] = converters_tmp[str(idx)]
            logger.debug('converters:', converters, type(converters))

        numeric_columns = list(result.columns[1:])
        logger.debug('path =', path)
        all_problems_on_data = IO.check_convertion_to_float(numeric_columns, result)
        idx: int = 0
        for key, value in converters.items():
            assert isinstance(value, type), f'Erro: {value} não é um tipo válido para coluna {key}'
            # Mapeia os tokens inválidos para NaN
            if (key in all_problems_on_data.keys()):
                result[key] = result[key].replace(to_replace=all_problems_on_data[key], value=np.nan)
            logger.debug('coluna:', key, '\ttipo:', value)
            if '<NA>' in result[key].values:
                result[key] = result[key].replace(to_replace='<NA>', value='NaN')
            # Faz a conversão de tipo para cada coluna
            column_series = result[key]
            logger.debug(column_series, type(column_series), type(column_series[0]))
            if value == pd.Timestamp or value == datetime:
                result[key] = pd.to_datetime(result[key], format='%Y-%m-%d %H:%M:%S')
            else:
                result[key] = result[key].astype(value)

        logger.debug(result.info())
        return result

    @staticmethod
    def dataframe_to_csv_file(df: pd.DataFrame, path: Path) -> None:
        # Se ocorrer erro devemos lançar Exception
        #
        # O Pandas salva por padrão datetime como neste exemplo: "2022-11-29 10:40:55".
        # Este é um formato de data e hora padrão ISO 8601 que é um padrão internacional
        # para representação de datas e horas, e é amplamente utilizado em sistemas de
        # computação e comunicação. O formato ISO 8601 para datas e horas é definido
        # como "AAAA-MM-DDTHH:MM:SS", onde "AAAA" representa o ano com quatro dígitos,
        # "MM" representa o mês com dois dígitos, "DD" representa o dia com dois dígitos,
        # "T" é o separador entre a data e a hora, "HH" representa a hora com dois dígitos,
        # "MM" representa os minutos com dois dígitos e "SS" representa os segundos com
        # dois dígitos. No formato "2022-11-29 10:40:55", a data e a hora são separadas
        # por um espaço em vez do caractere "T", mas ainda segue o padrão ISO 8601
        # para representação de datas e horas.
        # Veja: https://www.w3.org/TR/NOTE-datetime
        path_absolute = path.absolute()
        logger.debug(f'csv_file_path = {path_absolute}')
        df.to_csv(path, sep=',', na_rep='NaN', float_format=None,
            header=True, index=False, index_label=None, mode='w',
            encoding='utf-8', compression='infer', quoting=csv.QUOTE_MINIMAL,
            quotechar='"', lineterminator=os.linesep, chunksize=None,
            date_format=None, doublequote=True, escapechar=None,
            decimal='.', errors='strict', storage_options=None)

def __validate_type_convertion_test(path: Path):
    logger.debug('\n\n__validate_type_convertion(). Path:', path, '\n')
    column_types: list[type] = [pd.Timestamp, np.float32, np.float32]
    nan_values:list[str] = ['Bad', 'Configure']
    parse_dates: list[int] = [1]
    date_format:str = 'AAAA-MM-DD HH:MM:SS'
    df_whith_nans = IO.read_csv_file(path, column_types, nan_values, parse_dates, date_format)
    logger.debug('df_whith_nans:\n', df_whith_nans)
    logger.debug('Tipos das colunas no Dataframe da Série Temporal:')
    for col in df_whith_nans.columns:
        for value in df_whith_nans[col]:
            logger.debug(col, '\t', value, type(value))
        logger.debug('-----------------------------------------------')

if __name__ == "__main__":
    # Lê o arquivo CSV e gera um Dataframe com os tipos de dados corretos informados
    # na lista column_types que deve corresponder na mesma ordem com as colunas.
    __validate_type_convertion_test(Path('data/csv/ts_01.csv'))
    __validate_type_convertion_test(Path('data/csv/ts_02.csv'))



import os
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
from t8s import get_sample_df
from t8s.log_config import LogConfig
from t8s.util import Util
from t8s.io import IO
from t8s.ts import TimeSerie
from t8s.ts_writer import TSWriter, WriteParquetFile
from t8s.ts_builder import TSBuilder
from t8s.ts_builder import ReadParquetFile
from behave import given, when, then, use_step_matcher, step # type: ignore
from behave.model import Table
from behave_pandas import table_to_dataframe, dataframe_to_table
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL

LogConfig().initialize_logger(DEBUG)
logger = LogConfig().getLogger()

"""
Feature: Convert a multivariate Timeseries to list of univariate Timeseries and vice versa

Value Statement:
    As a data analyst
    I want the ability to convert between Timeseries types ['univariate', 'multivariate'] for use in different situations
    So I can start analyzing the data right away and come up with solutions for the business.

  Background:
    Given that I have a T8S_WORKSPACE_DIR and a long format time series persisted to a Parquet file
"""

@given(u'that I have a T8S_WORKSPACE_DIR and a long format time series persisted to a Parquet file')
def background(context):
    logger.info(u'STEP: Given that I have a T8S_WORKSPACE_DIR and a long format time series persisted to a Parquet file')

    if context.status == 'data directory empty':
        context.create_sample_ts_and_save_as_parquet(context)
        context.status = 'data directory with parquet file'

    # O método before_feature() em features/environment.py atualiza o contexto
    logger.info(f'-------------------------------------------------')
    logger.info(f'Background @given: T8S_WORKSPACE_DIR = {context.T8S_WORKSPACE_DIR}')
    logger.info(f'Background@given:  CSV_PATH = {context.CSV_PATH}')
    logger.info(f'Background@given:  PARQUET_PATH = {context.PARQUET_PATH}')
    context.list_files(f'Background@given:  ', context)
    # logger.info(f'\background : context.ts1 -> \n{str(context.ts1)}')
    logger.info(f'-------------------------------------------------')
    # A forma de passar estes dados para os steps seguintes é usando o objeto context

"""
  Scenario: Conversion of Timeseries types ['univariate', 'multivariate'] for use in different situations
    Given that I create a Timeseries using the selected parquet file in the T8S_WORKSPACE/data/parquet directory
    When I convert Timeseries from long format to wide format and check the convertion
    Then I can convert the Timeseries from multivariate to a list of univariate Timeseries
    And I convert the list of univariate Timeseries into a single multivariate Timeseries
    And I check the result.
    # Constraint: The Timeseries has no invalid values
"""

@given(u'that I create a Timeseries using the selected parquet file in the T8S_WORKSPACE/data/parquet directory')
def create_time_serie_from_parquet_file(context):
    file_name_of_time_series_in_long_format = 'ts_long_01.parquet'
    path_str: str = str(context.PARQUET_PATH) + '/' + file_name_of_time_series_in_long_format
    path = Path(path_str)
    logger.debug('path: ' + str(path))
    ctx = TSBuilder(ReadParquetFile())
    ts1: TimeSerie = ctx.build_from_file(Path(path_str))
    assert int(ts1.features) == 3
    assert ts1.format == 'long'
    assert len(ts1.df) == 8
    context.ts1 = ts1
    context.list_files(f'create_time_serie_from_parquet_file() \n', context)

@when(u'I convert Timeseries from long format to wide format and check the convertion')
def convert_time_serie_from_long_to_wide_format(context):
    logger.info(f'context.ts1 BEFORE -> \n{str(context.ts1.format)}')
    ts1: TimeSerie = context.ts1
    ts1.to_wide()
    logger.info(f'context.ts1 AFTER  -> \n{str(context.ts1.format)}\n{str(context.ts1.df.head())}')

@then(u'I can convert the Timeseries from multivariate to a list of univariate Timeseries')
def convert_time_serie_from_multivariate_to_list_of_univariate(context):
    # logger.info(f'context.ts1 BEFORE -> \n{str(context.ts1)}')
    pass

@then(u'I convert the list of univariate Timeseries into a single multivariate Timeseries')
def convert_list_of_univariate_to_multivariate(context):
    # logger.info(f'context.ts1 BEFORE -> \n{str(context.ts1)}')
    pass

@then(u'I check the result.')
def check_result(context):
    # logger.info(f'context.ts1 BEFORE -> \n{str(context.ts1)}')
    logger.info(f'split/join test passed')



# -*- coding: utf-8 -*-

from __future__ import annotations
from typing import TypeVar

T = TypeVar('T', bound='Node')


class Node:
    def __init__(self, value: int, children: list[T] = []) -> None:
        self.value = value
        self.children = children

    def add_child(self, child: T) -> None:
        self.children.append(child)  # type: ignore[arg-type]


if __name__ == '__main__':
    root = Node(1)
    child = Node(2)
    root.add_child(child)
    print(root.children[0].value)
